from PyQt5 import QtCore, QtGui, QtWidgets
import QT.gui
import QT.gui_report
import sys
import dice


# define globals
formula_log = [""]
formula_log_model = None
dice_log = []
dice_log_model = None
dice_tower = dice.Dice()

app = None
gui_main = None
MainWindow = None
gui_report = None
ReportWindow = None

def apply_ui_connections():
    """Overlay that connects up the GUI so that we can modularly replace the gui.py from QT5

    Args:
        gui_obj (gui.Ui_MainWindow): Main window GUI object
    """
    global gui_main, MainWindow, gui_report, reportWindow

    # set window icon
    # MainWindow.setWindowIcon(QtGui.QIcon('C:\Dropbox\_SCRIPTS\chipys-5e-companion\chipys-5e-tools\chipys_5e_tools\img\Chipy128.ico'))
    # MainWindow.setWindowIcon(QtGui.QIcon('C:\Dropbox\_SCRIPTS\chipys-5e-companion\chipys-5e-tools\chipys_5e_tools\img\ChipyLogo.png'))
    app.setWindowIcon(QtGui.QIcon('C:\Dropbox\_SCRIPTS\chipys-5e-companion\chipys-5e-tools\chipys_5e_tools\img\ChipyLogo.png'))
    MainWindow.setWindowTitle("Chipy's 5E Dice Sim")

    # link menus
    gui_main.actionExit.triggered.connect(lambda: app.exit())
    gui_main.actionProject_GitHub.triggered.connect(lambda: show_popup("Thanks for your curiosity! Please feel free to check out the project at https://github.com/iamchipy/chipys-5e-companion"))
    gui_main.actionChipy_Dev.triggered.connect(lambda: show_popup("Thanks for your curiosity! You can find more of my stuff at www.chipy.dev"))

    # link buttons
    gui_main.simulate.clicked.connect(lambda: run_sim(gui_main))
    gui_main.roll_active.clicked.connect(lambda: roll_active(gui_main))

    # connect listView log to click even with index
    gui_main.dice_log.clicked[QtCore.QModelIndex].connect(click_dice_log)
    gui_main.formula_log.clicked[QtCore.QModelIndex].connect(click_formula_log)

def show_popup( in_str):
    msg = QtWidgets.QMessageBox()
    msg.setWindowTitle("Results!")
    msg.setText(in_str)
    msg_instance = msg.exec_()  # exec with instanciation

def click_dice_log(index):
    global dice_log_model, gui_main
    item = dice_log_model.itemFromIndex(index)
    log_str = str(item.text())
    log_list = log_str.split("'")
    formula_used_was = log_list[1]
    gui_main.attack_formula.setText(formula_used_was)

def click_formula_log(index):
    global formula_log_model, gui_main
    item = formula_log_model.itemFromIndex(index)
    formula_used_was = str(item.text())
    gui_main.attack_formula.setText(formula_used_was)    

def roll_active(gui_obj):
    global dice_log, dice_log_model, formula_log, formula_log_model
    dice_tower = dice.Dice()
    formula = gui_obj.attack_formula.text()
    flag_adv = gui_obj.flag_adv.isChecked()
    flag_bls = gui_obj.flag_bls.isChecked()
    flag_dis = gui_obj.flag_dis.isChecked()
    flag_ela = gui_obj.flag_ela.isChecked()
    flag_ins = gui_obj.flag_ins.isChecked()
    flag_gwm = gui_obj.flag_gwm.isChecked()
    ac= gui_obj.armor_class.value()
    hit_rate = [0,0]
  

    roll, log, formula = dice_tower.r(formula, show_rolls=True,
                                                flag_adv = flag_adv,
                                                flag_bls = flag_bls,
                                                flag_dis = flag_dis,
                                                flag_ela = flag_ela,
                                                flag_ins = flag_ins,
                                                flag_gwm = flag_gwm)

    # build log
    dice_log.append([formula,roll,log])
    dice_log_model = QtGui.QStandardItemModel()
    gui_obj.dice_log.setModel(dice_log_model)
    dice_log_model.clear()
    for r in dice_log[-8:]:
        dice_log_model.appendRow(QtGui.QStandardItem(str(r)))
        # log hits
        if  int(r[1]) > ac:
            hit_rate[0] +=1
        hit_rate[1] +=1

    # build log last used formula
    expanded_var = formula_log[-1:][0]
    if formula != expanded_var:
        formula_log.append(formula)
    formula_log_model = QtGui.QStandardItemModel()
    gui_obj.formula_log.setModel(formula_log_model)
    formula_log_model.clear()
    for r in formula_log[-8:]:
        formula_log_model.appendRow(QtGui.QStandardItem(str(r)))        

    # build hit rate
    hit_rate = (100*hit_rate[0])//hit_rate[1]

    # display the current roll results
    gui_obj.result.setText(str(roll))
    gui_obj.hit_chance.setText(str(hit_rate))
    return roll
    
def run_sim(gui_obj):
    global dice_tower, ReportWindow, gui_report
    gui_obj.result.setText("Init...")
    formula = gui_obj.attack_formula.text()
    max_roll = dice_tower.max_roll(formula)
    roll_tally = [0 for i in range(max_roll)]
    hit_tally = [0 for i in range(35)]
    sim_count = gui_obj.sim_count.text()
    ac= gui_obj.armor_class.value()
    print(f"Starting {sim_count} itterations...")

    for i in range(int(sim_count)):
    # for i in range(1000):
        # roll dice
        rolled = roll_active(gui_obj)
        
        # log the roll in our tallies
        roll_tally[rolled-1] += 1
        if ac <0:
            for dc in range(len(hit_tally)):
                if rolled >= dc:
                    hit_tally[dc-1] +=1
        elif rolled >= ac:
            hit_tally[ac-1] += 1

    # build report

    roll_tally = roll_tally[1:]
    roll_tally_ratios = [round((roll_tally[i]/int(sim_count))*100,1) for i in range(max_roll-1)]
    print("How hard did you hit:")
    print(roll_tally)
    print(roll_tally_ratios)

    hit_tally = hit_tally[:-1]
    hit_tally_ratios = [round((hit_tally[i]/int(sim_count))*100,1) for i in range(len(hit_tally))]
    print("How often did you hit(at what AC):")
    print(hit_tally)
    print(hit_tally_ratios)    

    table_data = {  'roll_tally':roll_tally,
                    'roll_tally_ratios':roll_tally_ratios,
                    'hit_tally':hit_tally,
                    'hit_tally_ratios':hit_tally_ratios}

    # display
    gui_obj.result.setText("Done!")
    build_report_table(table_data, f"Results for {sim_count} itterations of {formula} VS an ArmorClass [{ac}]")
        
def build_report_table(table_data:dict, report_title:str="Results of simulation:"):
    global gui_report, ReportWindow

    largest_row = 0
    for name, set in table_data.items():
        print(len(set))
        print(set)
        if len(set) > largest_row:
            largest_row = len(set)

    gui_report.report_title.setText(report_title)
    ReportWindow.setWindowTitle("Chipy's 5E Dice Sim Report")

    gui_report.report_table.setRowCount(len(table_data))
    gui_report.report_table.setColumnCount(largest_row)

    # gui_report.report_table.setHorizontalHeaderLabels(("a","aa"))
    # gui_report.report_table.setItem(0,0,QtWidgets.QTableWidgetItem("test"))
    gui_report.report_table.setVerticalHeaderLabels(("roll","roll","hit","hit"))

    r=0
    for name, list in table_data.items():
        gui_report.report_table.setRowHeight(r,10)
        for c in range(len(list)):
            gui_report.report_table.setItem(r,c,QtWidgets.QTableWidgetItem(str(list[c])))
            gui_report.report_table.setColumnWidth(c,33)
            
        r+=1

    ReportWindow.show()

if __name__ == "__main__":
    import sys
    import ctypes

    # required for Windows to recognize a Python script as it's own applications and thus have a unique Taskbar Icon
    # https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
    myappid = u'chipy.5ETools.DiceSim.v0.1.4' # arbitrary string
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    # build main GUI
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    gui_main = QT.gui.Ui_MainWindow()
    gui_main.setupUi(MainWindow)
    MainWindow.show()

    # build report GUI
    ReportWindow = QtWidgets.QWidget()
    gui_report = QT.gui_report.Ui_Form()
    gui_report.setupUi(ReportWindow)

    # Modify the gui with connections and links
    apply_ui_connections()  # here we modify actions to the GUI

    # run app as the last thing in the script
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(517, 272)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setStatusTip("")
        self.centralwidget.setWhatsThis("")
        self.centralwidget.setObjectName("centralwidget")
        self.flag_adv = QtWidgets.QCheckBox(self.centralwidget)
        self.flag_adv.setEnabled(True)
        self.flag_adv.setGeometry(QtCore.QRect(10, 10, 81, 17))
        self.flag_adv.setObjectName("flag_adv")
        self.flag_gwm = QtWidgets.QCheckBox(self.centralwidget)
        self.flag_gwm.setEnabled(True)
        self.flag_gwm.setGeometry(QtCore.QRect(10, 70, 131, 17))
        self.flag_gwm.setObjectName("flag_gwm")
        self.roll_active = QtWidgets.QPushButton(self.centralwidget)
        self.roll_active.setGeometry(QtCore.QRect(210, 170, 81, 31))
        font = QtGui.QFont()
        font.setPointSize(16)
        font.setBold(True)
        font.setWeight(75)
        self.roll_active.setFont(font)
        self.roll_active.setObjectName("roll_active")
        self.armor_class = QtWidgets.QSpinBox(self.centralwidget)
        self.armor_class.setEnabled(True)
        self.armor_class.setGeometry(QtCore.QRect(10, 130, 49, 22))
        self.armor_class.setMinimum(-1)
        self.armor_class.setMaximum(35)
        self.armor_class.setProperty("value", -1)
        self.armor_class.setObjectName("armor_class")
        self.label = QtWidgets.QLabel(self.centralwidget)
        self.label.setEnabled(True)
        self.label.setGeometry(QtCore.QRect(60, 130, 68, 16))
        self.label.setObjectName("label")
        self.label_2 = QtWidgets.QLabel(self.centralwidget)
        self.label_2.setGeometry(QtCore.QRect(130, 180, 98, 20))
        self.label_2.setObjectName("label_2")
        self.flag_ins = QtWidgets.QCheckBox(self.centralwidget)
        self.flag_ins.setEnabled(True)
        self.flag_ins.setGeometry(QtCore.QRect(10, 90, 131, 17))
        self.flag_ins.setObjectName("flag_ins")
        self.flag_bls = QtWidgets.QCheckBox(self.centralwidget)
        self.flag_bls.setEnabled(True)
        self.flag_bls.setGeometry(QtCore.QRect(10, 110, 131, 17))
        self.flag_bls.setObjectName("flag_bls")
        self.sim_count = QtWidgets.QLineEdit(self.centralwidget)
        self.sim_count.setEnabled(True)
        self.sim_count.setGeometry(QtCore.QRect(12, 180, 101, 20))
        self.sim_count.setObjectName("sim_count")
        self.label_3 = QtWidgets.QLabel(self.centralwidget)
        self.label_3.setEnabled(True)
        self.label_3.setGeometry(QtCore.QRect(10, 160, 101, 16))
        self.label_3.setObjectName("label_3")
        self.flag_dis = QtWidgets.QCheckBox(self.centralwidget)
        self.flag_dis.setEnabled(True)
        self.flag_dis.setGeometry(QtCore.QRect(10, 30, 101, 17))
        self.flag_dis.setObjectName("flag_dis")
        self.attack_formula = QtWidgets.QLineEdit(self.centralwidget)
        self.attack_formula.setGeometry(QtCore.QRect(130, 210, 161, 20))
        self.attack_formula.setObjectName("attack_formula")
        self.flag_ela = QtWidgets.QCheckBox(self.centralwidget)
        self.flag_ela.setEnabled(True)
        self.flag_ela.setGeometry(QtCore.QRect(10, 50, 101, 17))
        self.flag_ela.setObjectName("flag_ela")
        self.result = QtWidgets.QLineEdit(self.centralwidget)
        self.result.setGeometry(QtCore.QRect(300, 180, 111, 51))
        font = QtGui.QFont()
        font.setFamily("Rubik SemiBold")
        font.setPointSize(36)
        font.setBold(True)
        font.setWeight(75)
        self.result.setFont(font)
        self.result.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.result.setObjectName("result")
        self.dice_log = QtWidgets.QListView(self.centralwidget)
        self.dice_log.setGeometry(QtCore.QRect(300, 10, 211, 151))
        self.dice_log.setObjectName("dice_log")
        self.formula_log = QtWidgets.QListView(self.centralwidget)
        self.formula_log.setGeometry(QtCore.QRect(140, 10, 151, 151))
        self.formula_log.setObjectName("formula_log")
        self.simulate = QtWidgets.QPushButton(self.centralwidget)
        self.simulate.setEnabled(True)
        self.simulate.setGeometry(QtCore.QRect(10, 200, 101, 31))
        font = QtGui.QFont()
        font.setPointSize(16)
        self.simulate.setFont(font)
        self.simulate.setObjectName("simulate")
        self.hit_chance = QtWidgets.QLineEdit(self.centralwidget)
        self.hit_chance.setGeometry(QtCore.QRect(420, 180, 91, 51))
        font = QtGui.QFont()
        font.setFamily("Rubik SemiBold")
        font.setPointSize(36)
        font.setBold(True)
        font.setWeight(75)
        self.hit_chance.setFont(font)
        self.hit_chance.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.hit_chance.setObjectName("hit_chance")
        self.label_4 = QtWidgets.QLabel(self.centralwidget)
        self.label_4.setGeometry(QtCore.QRect(300, 160, 98, 20))
        self.label_4.setObjectName("label_4")
        self.label_5 = QtWidgets.QLabel(self.centralwidget)
        self.label_5.setGeometry(QtCore.QRect(430, 160, 98, 20))
        self.label_5.setObjectName("label_5")
        self.flag_adv.raise_()
        self.flag_gwm.raise_()
        self.armor_class.raise_()
        self.label.raise_()
        self.label_2.raise_()
        self.flag_ins.raise_()
        self.flag_bls.raise_()
        self.sim_count.raise_()
        self.label_3.raise_()
        self.flag_dis.raise_()
        self.attack_formula.raise_()
        self.flag_ela.raise_()
        self.result.raise_()
        self.dice_log.raise_()
        self.formula_log.raise_()
        self.roll_active.raise_()
        self.simulate.raise_()
        self.hit_chance.raise_()
        self.label_4.raise_()
        self.label_5.raise_()
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 517, 21))
        self.menubar.setObjectName("menubar")
        self.menuModes = QtWidgets.QMenu(self.menubar)
        self.menuModes.setObjectName("menuModes")
        self.menuAbout = QtWidgets.QMenu(self.menubar)
        self.menuAbout.setObjectName("menuAbout")
        self.menuConfigurations = QtWidgets.QMenu(self.menubar)
        self.menuConfigurations.setEnabled(False)
        self.menuConfigurations.setObjectName("menuConfigurations")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionHit_Chance = QtWidgets.QAction(MainWindow)
        self.actionHit_Chance.setObjectName("actionHit_Chance")
        self.actionExit = QtWidgets.QAction(MainWindow)
        self.actionExit.setObjectName("actionExit")
        self.actionProject_GitHub = QtWidgets.QAction(MainWindow)
        self.actionProject_GitHub.setObjectName("actionProject_GitHub")
        self.actionChipy_Dev = QtWidgets.QAction(MainWindow)
        self.actionChipy_Dev.setObjectName("actionChipy_Dev")
        self.menuModes.addAction(self.actionHit_Chance)
        self.menuModes.addSeparator()
        self.menuModes.addAction(self.actionExit)
        self.menuAbout.addAction(self.actionProject_GitHub)
        self.menuAbout.addAction(self.actionChipy_Dev)
        self.menubar.addAction(self.menuModes.menuAction())
        self.menubar.addAction(self.menuConfigurations.menuAction())
        self.menubar.addAction(self.menuAbout.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.flag_adv.setStatusTip(_translate("MainWindow", "Toggle to roll with Advantage"))
        self.flag_adv.setText(_translate("MainWindow", "Advantage"))
        self.flag_gwm.setStatusTip(_translate("MainWindow", "Toggle to roll with the -5 hit for Great Weapon Master & SharpShooter"))
        self.flag_gwm.setText(_translate("MainWindow", "GWM/SharpShooter"))
        self.roll_active.setText(_translate("MainWindow", "Roll"))
        self.armor_class.setToolTip(_translate("MainWindow", "Select Armor class (or leave on \'-1\' to simular for all classes)"))
        self.armor_class.setStatusTip(_translate("MainWindow", "Select Armor class (or leave on \'-1\' to simular for all classes)"))
        self.label.setText(_translate("MainWindow", "Foe\'s AC"))
        self.label_2.setText(_translate("MainWindow", "Attack Formula"))
        self.flag_ins.setStatusTip(_translate("MainWindow", "Toggle to roll with Ispiration (Advantage)"))
        self.flag_ins.setText(_translate("MainWindow", "Inpiration"))
        self.flag_bls.setStatusTip(_translate("MainWindow", "Toggle to roll with as blessed (+1d4)"))
        self.flag_bls.setText(_translate("MainWindow", "Blessed"))
        self.sim_count.setStatusTip(_translate("MainWindow", "Number of dice roll attempts to simulate"))
        self.sim_count.setText(_translate("MainWindow", "9001"))
        self.label_3.setText(_translate("MainWindow", "Simulation count"))
        self.flag_dis.setStatusTip(_translate("MainWindow", "Toggle to roll with Disadvantage"))
        self.flag_dis.setText(_translate("MainWindow", "Disadvantage"))
        self.attack_formula.setStatusTip(_translate("MainWindow", "Dice Attack formula (your hit dice when attacking) Example: 1d20+1d4+2+1"))
        self.attack_formula.setText(_translate("MainWindow", "1d20+1"))
        self.flag_ela.setStatusTip(_translate("MainWindow", "Toggle to roll with \"double\" advantage when you have Elven Accuracy"))
        self.flag_ela.setText(_translate("MainWindow", "Elven Accuracy"))
        self.result.setStatusTip(_translate("MainWindow", "Final value of the last roll"))
        self.result.setText(_translate("MainWindow", "###"))
        self.dice_log.setStatusTip(_translate("MainWindow", "Quick load formula from recent rolls"))
        self.formula_log.setStatusTip(_translate("MainWindow", "Quick recent roll formulas"))
        self.simulate.setStatusTip(_translate("MainWindow", "Start simulation of current roll configuration"))
        self.simulate.setText(_translate("MainWindow", "Simlulate"))
        self.hit_chance.setStatusTip(_translate("MainWindow", "Avarage hit rate vs current set AC (over the last 8 rolls)"))
        self.hit_chance.setText(_translate("MainWindow", "##"))
        self.label_4.setText(_translate("MainWindow", "Last roll:"))
        self.label_5.setText(_translate("MainWindow", "Hit Rate: "))
        self.menuModes.setTitle(_translate("MainWindow", "Modes"))
        self.menuAbout.setTitle(_translate("MainWindow", "About"))
        self.menuConfigurations.setTitle(_translate("MainWindow", "Configurations"))
        self.actionHit_Chance.setText(_translate("MainWindow", "Hit Chance"))
        self.actionExit.setText(_translate("MainWindow", "E&xit"))
        self.actionProject_GitHub.setText(_translate("MainWindow", "Project GitHub"))
        self.actionChipy_Dev.setText(_translate("MainWindow", "Chipy.Dev"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    MainWindow = QtWidgets.QMainWindow()
    ui = Ui_MainWindow()
    ui.setupUi(MainWindow)
    MainWindow.show()
    sys.exit(app.exec_())



import random as rng

def __init__():
    pass

class Dice:
    """General dice object to collect and run dice formulas 
    """
    def __init__(self):
        pass

    def r(self, in_str:str="1d20", flag_adv:bool=False, show_rolls:bool= False, flag_bls:bool= False,flag_dis:bool= False,flag_ela:bool= False,flag_ins:bool= False,flag_gwm:bool= False, flag_spec:int=0):
        """"alias for self.formula()"""
        return self.formula(in_str, flag_adv, show_rolls, flag_bls, flag_dis, flag_ela, flag_ins, flag_gwm, flag_spec)

    def f(self, in_str:str="1d20", flag_adv:bool=False, show_rolls:bool= False,flag_bls:bool= False,flag_dis:bool= False,flag_ela:bool= False,flag_ins:bool= False,flag_gwm:bool= False, flag_spec:int=0):
        """"alias for self.formula()"""
        return self.formula(in_str, flag_adv, show_rolls, flag_bls, flag_dis, flag_ela, flag_ins, flag_gwm, flag_spec)

    def formula(self, in_str:str="1d20", flag_adv:bool=False, show_rolls:bool= False, flag_bls:bool= False,flag_dis:bool= False,flag_ela:bool= False,flag_ins:bool= False,flag_gwm:bool= False, flag_spec:int=0):
        """_summary_

        Args:
            in_str (str, optional): Desired dice-formula to roll written in the #d##+#d##+# format. EXAMPLE: 1d20+2d4+3 Defaults to "1d20".
            advantage (bool, optional): When True the first dice in the formula will get one additional roll and the lower will be dropped. So "2d20" would become 3x d20 rolls with the lowest dropped. Defaults to False.
            show_rolls (bool, optional): When True return value will be an array including each dice roll and the originating formula. Defaults to False.

        Returns:
            int or list: depending on "show_rolls" flag will either return a single int value or a list [value,["each","dice","roll"],"formula given"]
        """
        dice_report= ""
        adv_rolled = False
        int_value = 0
        formula = in_str.split("+")  # | (pipe) separates delimiters
        # Blessed gets handled here ------------- flag_bls
        if flag_bls:
            formula.append("1d4")
            
        for element in formula:
            if "d" in element:
                # Advantage get handled here ------------- flag_adv
                # Disadvantage get handled here ------------- flag_dis
                # ElvenAccuracy get handled here ------------- flag_ela
                # Inspiration get handled here ------------- flag_ins

                # all of these are only being applied to the first dice in the formula and only once 
                if not adv_rolled:
                    i, s = self.roll(element, flag_adv=flag_adv, flag_dis=flag_dis, flag_ins=flag_ins, flag_ela=flag_ela)
                    dice_report += str(s)
                    int_value += i
                    adv_rolled = True
                else:
                    i, s = self.roll(element)
                    dice_report += str(s)
                    int_value += i                    
            else:
                int_value += int(element)

        # SharpShooter/GreatWepMaster gets handled here ------------- flag_gwm
        if flag_gwm:
            int_value -= 5

        # SPECIAL gets handled here ------------- flag_gwm
        int_value += flag_spec            
        
        # check what info to return and return it
        if show_rolls:
            return int_value, dice_report, in_str
        else:
            return int_value
    
    def roll(self, dice_string="1d20", flag_adv:bool=False, flag_dis:bool=False, flag_ins:bool=False, flag_ela:bool=False)-> int:
        """_summary_

        Args:
            dice_string (str, optional): Dice formula. Defaults to "1d20".
            flag_adv (bool, optional): A toggle indicating if there should be advantage (it'll roll one extra dice and drop lowest). Defaults to False.
            flag_dis (bool, optional): A toggle indicating if there should be disadvantage (it'll roll one extra dice and drop highest). Defaults to False.
            flag_ins (bool, optional): A toggle indicating if there should be advantage (it'll roll one extra dice and drop lowest). Defaults to False.
            flag_ela (bool, optional): A toggle to roll a unique stacking advantage. Defaults to False.

        Returns:
            int: dice roll
        """
        # calc advantange
        adv_counter = 0
        if flag_adv or flag_ins:
            adv_counter +=1
        if flag_dis:
            adv_counter -=1          

        adv_rolled = False
        dice_rolls = []
        kept_rolls = []
        dice_array = dice_string.split("d")
        # loop ones for each dice being roll 
        for i in range(int(dice_array[0])):
            # base dice
            roll = self._roll_value(dice_array[1])
            dice_rolls.append(roll)
            kept_rolls.append(roll)      

            # if we've not yet rolled advantages for this dice formula yet (can't only get one adv)
            # we do this on the first roll so that we can manipulate the list safely
            if not adv_rolled:
                # if there are adv/dis to be applied roll a second dice
                if adv_counter!=0:
                    dice_rolls.append(self._roll_value(dice_array[1]))
                    adv_rolled = True
                    # if we have positive advantage then drop lower of the two
                    if adv_counter>0:
                        
                        # if we have Elven Accuracy we do a 3rd dice and drop lowest
                        if flag_ela:
                            # roll the double adv and keep best
                            dice_rolls.append(self._roll_value(dice_array[1]))
                            dice_rolls.sort()
                            kept_rolls = dice_rolls[2:]
                        else:
                            # sort and keep best per normal adv
                            dice_rolls.sort()
                            kept_rolls = dice_rolls[1:]
                    else:
                        dice_rolls.sort()
                        kept_rolls = dice_rolls[:1]      

        return sum(kept_rolls), dice_rolls

    def _roll_value(self, sides_to_roll=0):
        """Most basic roll command to generate a random number within the range

        Returns:
            int: face value of this roll
        """
        if not sides_to_roll:
            sides_to_roll = self.sides
        return int(rng.randrange(1,int(sides_to_roll)+1))

    def max_roll(self, formula):
        parts= formula.split("+")
        max = 0
        for item in parts:
            if "d" in item:
                d = item.split("d")
                max+=(int(d[0])*int(d[1]))
            else:
                max+=int(item)
        return max


if __name__ == "__main__":
    d = Dice()
    # print("1d20 ", d.r("1d20"))
    # print("1d20+1 ", d.r("1d20+1"))
    # print("2d20 ", d.r("2d20"))
    # print("2d20 a", d.r("2d20",True))
    # print("2d20 a", d.r("2d20",True,True))
    # print("2d20 a", d.r("4d20",show_rolls=True, flag_adv=1, flag_ela=1))
    # print("1d20 ", d.r("1d20",show_rolls=1,flag_gwm=1))
    # print("1d20 ", d.r("1d20",show_rolls=1,flag_spec=100))
    print("2d20 ", d.r("2d20",show_rolls=1,flag_adv=True))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'gui_report.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(1035, 184)
        self.report_table = QtWidgets.QTableWidget(Form)
        self.report_table.setGeometry(QtCore.QRect(0, 50, 1031, 131))
        self.report_table.setObjectName("report_table")
        self.report_table.setColumnCount(0)
        self.report_table.setRowCount(0)
        self.report_title = QtWidgets.QLabel(Form)
        self.report_title.setGeometry(QtCore.QRect(10, 0, 1021, 41))
        font = QtGui.QFont()
        font.setFamily("Yu Gothic UI Semibold")
        font.setPointSize(28)
        font.setBold(True)
        font.setWeight(75)
        self.report_title.setFont(font)
        self.report_title.setObjectName("report_title")

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.report_title.setText(_translate("Form", "TextLabel"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()
    ui = Ui_Form()
    ui.setupUi(Form)
    Form.show()
    sys.exit(app.exec_())
